import traceback
from functools import wraps
from logging import getLogger
from typing import Type, Optional, Union, Callable, Tuple, Sequence

from flask import request, jsonify
from marshmallow import Schema
from marshmallow import ValidationError
from sqlalchemy.exc import DatabaseError
from werkzeug import Response, exceptions  # noqa: PyPackageRequirements

from avatar_utils.core import create_response
from avatar_utils.validation.constants import ServiceHTTPCodes

logger = getLogger(__name__)

DEFAULT_VALIDATE_METHOD = 'GET'


def validate_json(request_schema: Optional[Type[Schema]] = None,
                  response_schema: Optional[Type[Schema]] = None,
                  methods: Union[str, Sequence[str]] = DEFAULT_VALIDATE_METHOD) -> Callable:
    if isinstance(methods, str):
        methods = {methods}

    methods = set(m.upper().strip() for m in methods)

    def decorator(f) -> Callable:

        @wraps(f)
        def wrapper(*args, **kw) -> Union[str, Tuple[Response, int]]:

            log_prefix = f'[ VALIDATION | {request.remote_addr} {request.method} > {request.url_rule} ]'
            step = 'pre processing'

            try:
                # skip decorator if unexpected HTTP method of request
                if request.method not in methods:
                    logger.debug(f'{log_prefix} skip {request.method} method, validator expect {methods}')
                    return f(*args, **kw)

                # validate request data
                step = 'validate request data'
                if request_schema:
                    validation_schema = request_schema()

                    if not request.is_json:
                        logger.warning(f'{log_prefix} < invalid content-type header')
                        return create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                               message='invalid content-type header')

                    validation_schema.load(request.json)
                    logger.debug(f'{log_prefix} request body validated')

                # call HTTP handler
                step = 'call HTTP handler'
                status = 200
                result = f(*args, **kw)
                if isinstance(result, tuple):
                    response, status = result
                else:
                    response = result

                if not isinstance(response, Response):
                    response = jsonify(result)

                # validate response data
                step = 'validate response data'
                if response_schema:
                    validation_schema = response_schema()

                    if response.status_code != ServiceHTTPCodes.OK.value:
                        return response, status

                    data = validation_schema.load(response.json)
                    logger.debug(f'{log_prefix} response validated')
                    response = jsonify(validation_schema.dump(data))

            except exceptions.BadRequest as err:
                logger.warning(f'{log_prefix} < Step: {step}. Exception: {err}')
                response, status = create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                                   message='invalid requests body')
            except ValidationError as err:
                logger.warning(f'{log_prefix} < Step: {step}. Exception: {err}')
                response, status = create_response(status=ServiceHTTPCodes.UNPROCESSABLE_ENTITY.value,
                                                   message='validation error',
                                                   data=dict(err.messages))
            except NotImplementedError as err:
                logger.warning(f'{log_prefix} < Step: {step}. Exception: {err}')
                response, status = create_response(status=ServiceHTTPCodes.NOT_IMPLEMENTED.value,
                                                   message=type(err).__name__,
                                                   data=dict(traceback=str(traceback.format_exc())))
            except DatabaseError as err:
                logger.warning(f'{log_prefix} < Step: {step}. Exception: {err}')
                try:
                    from app import db
                    logger.debug('db imported from app')
                    db.session.rollback()
                except ImportError:
                    logger.info('Cannot import db from app. Skip.')
                response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                   message=type(err).__name__,
                                                   data=dict(traceback=str(traceback.format_exc())))
            except (TypeError, Exception) as err:  # noqa
                logger.warning(f'{log_prefix} < Step: {step}. Exception: {err}')
                response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                   message=type(err).__name__,
                                                   data=dict(traceback=str(traceback.format_exc())))

            return response, status

        return wrapper

    return decorator



import json
from datetime import datetime
from logging import getLogger

from flask import request

from app import db
from avatar_utils.db.mixins import CRUDMixin, TimeCastMixin

logger = getLogger(__name__)


def format_text(text):
    return ' '.join(text.strip().split())


class Log(CRUDMixin, db.Model):
    __tablename__ = 'logs'

    id = db.Column(db.Integer(), primary_key=True)
    route = db.Column(db.String(), nullable=False)
    method = db.Column(db.String(10), nullable=False)
    request = db.Column(db.Text(), nullable=False)
    response = db.Column(db.Text(), nullable=True)
    ts = db.Column(db.DateTime(timezone=True), nullable=False, default=TimeCastMixin.utc_now)
    duration = db.Column(db.Float(), nullable=True)
    status_code = db.Column(db.String, nullable=True)
    message = db.Column(db.String, nullable=True)

    def __repr__(self):
        return """<log object>
        id: {}
        route: {}
        method: {}
        request: {}
        response: {}
        ts: {}
        duration: {}
        status_code: {}
        message: {}""".format(self.id, self.route, self.method, self.request, self.response, self.ts,
                              self.duration, self.status_code, self.message)

    @staticmethod
    def init(request):
        data = request.json
        log = Log.create(route=request.path, method=request.method, request=json.dumps(data))
        return log

    @staticmethod
    def complete(log, response):
        if log:
            log.update(duration=(datetime.utcnow() - log.ts).total_seconds(), status_code=response.status)
            try:
                data = response.json
                log.update(response=json.dumps(data), message=data.get('message'))
            except TypeError as err:
                log_prefix = f'[ EVENT LOGGING | {request.remote_addr} {request.method} > {request.url_rule} ]'
                logger.warning(f'{log_prefix} < Response is not in JSON format')
                logger.warning(
                    f'{log_prefix} < Message: {err.__class__.__name__}. Errors: {list(str(e) for e in err.args)}')
            db.session.commit()
            return True
        return False



import copy
import logging

import flask
import requests

from avatar_utils.tests.AbstractApp import AbstractApp
from avatar_utils.tests.Env import *

logger = logging.getLogger(__name__)


class RemoteApp(AbstractApp):
    client = requests

    # default env, specified once per whole test class
    default_env = DevEnv()

    # current env, could be changed in every test fuction
    current_env = copy.deepcopy(default_env)

    @property
    def base_url(self):
        return self.current_env.base_url

    def post(self, url, data=dict()) -> flask.Response:
        res = super().post(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    def get(self, url, data=dict()) -> flask.Response:
        res = super().get(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    # TODO transfer headers
    @staticmethod
    def convert_to_flask_response(response: requests.Response) -> flask.Response:
        res = flask.Response(
            response=response.text,
            status=response.status_code,
            mimetype=response.headers['content-type']
        )

        return res



import json
import re
from copy import deepcopy
from dataclasses import dataclass as python_dataclass
from logging import getLogger
from typing import Any

# import avatar_utils.objects - required
import avatar_utils.objects

logger = getLogger(__name__)


@python_dataclass
class Serializable:

    def __post_init__(self):
        self.repr_type = self.__fullname()

    # reliable but not the most productive way to serialize as a dict
    def to_dict(self) -> dict:
        json_data = json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

        return json.loads(json_data)

    @staticmethod
    def class_attributes(cls):
        attributes = []

        for key in cls.__dict__.keys():
            if key == '__dataclass_fields__':
                dataclass_fields = getattr(cls, key)
                for field in dataclass_fields.keys():
                    attributes.append(field)

        return attributes

    @staticmethod
    def any_to_dict(data: Any):
        data_copy = deepcopy(data)

        if isinstance(data_copy, Serializable):
            res = data_copy.to_dict()
            return res
        elif isinstance(data_copy, list):
            for i in range(len(data_copy)):
                data_copy[i] = Serializable.any_to_dict(data=data_copy[i])
            return data_copy
        elif isinstance(data_copy, dict):
            for key, value in data_copy.items():
                data_copy[key] = Serializable.any_to_dict(value)
            return data_copy
        else:
            raise NotImplementedError

    # все вложенные объекты десериализуются как Serializable, то есть у них должно быть поле repr_type
    @classmethod
    def from_dict(cls, data: Any, objects_existing_check: bool = False) -> Any:

        def from_dict_with_check(cls, data: Any) -> (Any, bool):
            if isinstance(data, dict):
                # serializable class
                if cls.__name__ == 'Serializable':
                    logger.debug('Deserialize object in general. Using Serializable')
                    repr_type = data.pop('repr_type', None)

                    if repr_type:
                        # check precisely
                        if not injection_detected(cls=cls, repr_type=repr_type):
                            # evaluate actual class
                            actual_cls = eval(repr_type)
                            return extract_class(actual_cls, data)
                        else:
                            logger.error('Injection detected. Skip evaluation')
                            return {}, False

                    else:
                        logger.debug('Dict is received')

                        result = {}
                        object_has_met = False
                        for k, v in data.items():
                            result[k], inner_object_has_met = from_dict_with_check(cls=Serializable, data=v)
                            object_has_met = object_has_met | inner_object_has_met

                        return result, object_has_met
                else:
                    # concrete class received
                    return extract_class(cls, data)

            elif isinstance(data, list):
                object_has_met = False
                i: int
                for i in range(data.__len__()):
                    data[i], inner_object_has_met = from_dict_with_check(cls, data[i])
                    object_has_met = object_has_met | inner_object_has_met
                return data, object_has_met
            else:
                return data, False

        def injection_detected(cls, repr_type: str):
            module_name = 'avatar_utils.objects'
            import sys, inspect
            for name, obj in inspect.getmembers(sys.modules[module_name]):
                if inspect.isclass(obj):
                    qualify_path = str(obj)[8:-2]
                    # concrete classes in avatar_utils.objects module
                    if qualify_path.startswith(module_name) and not qualify_path.lower().__contains__('abstract'):
                        # class exists
                        if qualify_path.endswith(repr_type):
                            # acceptable only
                            if re.match('^[A-Za-z_\\.]*$', qualify_path):
                                return False

            return True

        def extract_class(cls, data: Any):

            attributes = Serializable.class_attributes(cls)

            result = {}
            for k, v in data.items():
                if k in attributes:
                    result[k], skip = from_dict_with_check(cls=Serializable, data=v)
                else:
                    logger.warning('Skip unexpected keyword argument "%s"', k)
            extracted_cls = cls(**result)
            return extracted_cls, True

        result, object_has_met = from_dict_with_check(cls, deepcopy(data))

        if objects_existing_check:
            if not object_has_met:
                raise TypeError('No object has met')

        return result

    def __fullname(self):
        # o.__module__ + "." + o.__class__.__qualname__ is an example in
        # this context of H.L. Mencken's "neat, plausible, and wrong."
        # Python makes no guarantees as to whether the __module__ special
        # attribute is defined, so we take a more circumspect approach.
        # Alas, the module name is explicitly excluded from __qualname__
        # in Python 3.

        module = self.__class__.__module__
        if module is None or module == str.__class__.__module__:
            return self.__class__.__name__  # Avoid reporting __builtin__
        else:
            return module + '.' + self.__class__.__name__



from logging import getLogger

logger = getLogger(__name__)


def read_version(path):
    try:
        with open(path, encoding='utf-8') as f:
            version = f.readline().strip()
            logger.info('Current version: %s', version)

            if version.count('.') != 2:
                logger.warning('Version number is not in "major.minor.fix" format')

            return version
    except FileNotFoundError as err:
        logger.warning(err)

    return None
